open Ast
open Sast
open Env

module StringMap = Map.Make(String)

let classIndices: (string, int) Hashtbl.t =  Hashtbl.create 10

let createClassIndices cdecls=  
  let classHandler index cdecl=
    Hashtbl.add classIndices cdecl.cname index in (*scope handling is missing*)
  List.iteri classHandler cdecls

let builtinMethods = ["print"; "println"]

let isMain f = f.sfname = "main"

let get_methods l classy =  List.concat [classy.scbody.smethods;l]

let get_main m = try (List.hd (List.filter isMain (List.fold_left get_methods [] m))) with Failure("hd") -> raise(Failure("No main method was defined"))

let get_methods_minus_main m = snd (List.partition isMain (List.fold_left get_methods [] m))

let newEnv env = {
  envClassName = env.envClassName;
  envClassMaps = env.envClassMaps;
  envClassMap  = env.envClassMap;
  envLocals    = env.envLocals;
  envParams    = env.envParams;
  envReturnType = env.envReturnType;
  envBuiltinMethods = env.envBuiltinMethods;
}

let typOfSexpr = function
    SLiteral(i)			-> Int	
  | 	SBoolLiteral(b)			-> Bool	
  | 	SFLiteral(f)			-> Float
  | 	SStrLiteral(s) 			-> String
  | 	SChLiteral(c) 			-> Char
  | 	SId(_, d) 			-> d
  | 	SBinop(_, _, _, d) 		-> d
  |   SAssign(_, _, d) 		-> d
  | 	SCall(_, _, d)	-> d
  |  	SUnop(_, _, d) 			-> d 
  | 	SNewObject(_,_,d)	-> d
  | 	SAccessMember(_,_,d) -> d
  | _ -> Void

let rec typOfExpr = function
    Literal(i)			-> Int	
  | 	BoolLiteral(b)			-> Bool	
  | 	FLiteral(f)			-> Float
  | 	StrLiteral(s) 			-> String
  | 	ChLiteral(c) 			-> Char
  | 	Id(_) 			-> Int
  | 	Binop(_, _, e) 		-> typOfExpr e
  |   Assign(_, e) 		-> typOfExpr e
  | 	Call(_, e)	-> Int
  |   Unop(_, e) 			-> typOfExpr e
  | 	NewObject(s,_)	-> Object(s)
  | 	AccessMember(_,e) -> typOfExpr e
  | _ -> Void

let convertToSast classes main =

  let convertFormalToSast formal env =
    {
      sfvtyp = formal.fvtyp;
      sfvname = formal.fvname;
    }
  in
  let checkBinop e1 op e2 =
    ()
  in
  let checkUnop op e env =
    ()
  in
  let checkFuncCall s el env = 
    ()
  in
  let checkAssign e1 e2 env =
    ()  
  in
  let rec convertExprToSast expr env = match expr with
      Literal(i)	-> SLiteral(i)
    | BoolLiteral(b)	-> SBoolLiteral(b)
    | FLiteral(f)	-> SFLiteral(f)
    | StrLiteral(s)	-> SStrLiteral(s)
    | ChLiteral(c)	-> SChLiteral(c)
    | Null		-> SNull
    | Noexpr	-> SNoexpr
    | Id(id)	-> SId(id, Int (*getIdType id env*)) 
    | Binop(expr1, op, expr2)	-> ignore(checkBinop expr1 op expr2);SBinop(convertExprToSast expr1 env, op, convertExprToSast expr2 env, Int (*getType expr1 env)*))
    | Assign(e1, e2)		-> ignore(checkAssign e1 e2 env); SAssign(convertExprToSast e1 env, convertExprToSast e2 env,Int(* getType e1 env*))
    | Call(s, el)		-> ignore(checkFuncCall s el env); SCall(s, (List.map (fun e -> convertExprToSast e env) el), Int)
    | Unop(op, expr)		-> ignore(checkUnop op expr env); SUnop(op, convertExprToSast expr env, Int (*getType expr env*))
    | NewObject(s,el)	-> SNewObject(s, (List.map (fun e -> convertExprToSast e env) el), Int (*Object(s)*))
    | AccessMember(e1,e2) -> SAccessMember(convertExprToSast e1 env, convertExprToSast e2 env, Int (*getType e1 env*)) (* @TODO Double check type *)
  in
  let convertVdeclToSast vdecl env = {
    svaccess = vdecl.vaccess;
    svtyp = vdecl.vtyp;
    svname = vdecl.vname;
    svexpr = convertExprToSast vdecl.vexpr env;
  } in
  let checkLocalVariable dt id e env =
    env.envLocals <- StringMap.add id dt env.envLocals 
  in
  let checkReturn e env =
    ()
  in
  let checkIf e s1 s2 env =
    ()	
  in
  let checkFor e1 e2 e3 s env = 
    ()
  in
  let checkWhile e s env =
    ()
  in	
  let rec convertStmtToSast stmt env = match stmt with
      Block(sl)			-> SBlock(List.map (fun s -> convertStmtToSast s env) sl)
    | Expr(expr)			-> SExpr((convertExprToSast expr env), Int)
    | Variable(vdecl)		-> SVariable(convertVdeclToSast vdecl env)
    | LocalVariable(dt, id, expr)	-> (checkLocalVariable dt id expr env; SLocalVariable(dt, id, convertExprToSast expr env))
    | Return(expr)  		-> ignore(checkReturn expr env); SReturn((convertExprToSast expr env), Int)
    | If(expr, stmt1, stmt2)	-> ignore(checkIf expr stmt1 stmt2 env); SIf(convertExprToSast expr env, convertStmtToSast stmt1 (newEnv env), convertStmtToSast stmt2 (newEnv env))
    | For(expr1, expr2, expr3, stmt)-> ignore(checkFor expr1 expr2 expr3 stmt env); SFor(convertExprToSast expr1 env, convertExprToSast expr2 env, convertExprToSast expr3 env, convertStmtToSast stmt (newEnv env))
    | While(expr, stmt)		-> ignore(checkWhile expr stmt env); SWhile(convertExprToSast expr env, convertStmtToSast stmt (newEnv env))
  in

  (* Semantic Checking for class methods *)
  let rec strOfFormals fl = match fl with
    |  [] -> ""
    | h::t -> string_of_typ h ^ (strOfFormals t)
  in
  let getListOfFormalTypes c = List.map (fun f -> f.fvtyp) c.fformals 
  in
  let hasDuplicateFormalNames l = 
    let result = ref false
    in let names = List.map (fun f -> f.fvname) l
    in List.iter (fun e -> result := !result || e) (List.map (fun n -> List.length (List.filter (fun s -> s = n) names) > 1) names);!result
  in
  let checkMethod func_decl classEnv =
    {
      formalAccess = Public;
      formalName = "";
      formalTyps = [];
      formalTyp = Int;
    }

  in
  let setEnvParams formals env =
    List.map (fun f -> env.envParams <- StringMap.add f.fvname f.fvtyp env.envParams) formals

  in
  let convertMethodToSast func_decl classEnv =

    let methodSignature = checkMethod func_decl classEnv
    in
    let _ = classEnv.classMap.methodMap <- StringMap.add func_decl.fname methodSignature classEnv.classMap.methodMap
    in
    let env = {
      envClassName = classEnv.className;
      envClassMaps = classEnv.classMaps;
      envClassMap  = classEnv.classMap;
      envLocals    = StringMap.empty;
      envParams    = StringMap.empty;
      envReturnType= func_decl.ftyp;
      envBuiltinMethods = classEnv.builtinMethods;
    } in 
    let _ = setEnvParams func_decl.fformals env
    in 
    {
      sfaccess = func_decl.faccess;
      sfname = func_decl.fname;
      sfformals = List.map (fun f -> convertFormalToSast f env) func_decl.fformals; 
      sftyp = func_decl.ftyp;
      sfbody = List.map (fun s -> convertStmtToSast s env) func_decl.fbody;
    }

  in
  (* Semantic checking for class constructor *)
  let checkConstructor constructor classEnv = 
    (*		let formalTypes = getListOfFormalTypes constructor
      		in
      		let checking =

      			if(StringMap.mem (strOfFormals formalTypes) classEnv.classMap.constructorMap)
      				then raise (Failure("Duplicate Constructor Definition"))
      			else if ((List.length formalTypes <> 0) && (hasDuplicateFormalNames constructor.fformals = true))
      				then raise(Failure("Formal names must be unique"))  
      		in checking;
      	        {
                    	mscope = constructor.fscope;
                            mname  = constructor.fname;
                            mformalTypes = formalTypes;
      			mReturn = JVoid;
                    }
    *)
    {
      formalAccess = Public;
      formalName = "";
      formalTyps = [];
      formalTyp = Int;
    }
  in
  let convertConstructorToSast constructor classEnv =
    let constructorSignature = checkConstructor constructor classEnv
    in
    let _ = classEnv.classMap.constructorMap <- StringMap.add (strOfFormals constructorSignature.formalTyps) constructorSignature classEnv.classMap.constructorMap
    in
    let env = {
      envClassName = classEnv.className;
      envClassMaps = classEnv.classMaps;
      envClassMap  = classEnv.classMap;
      envLocals    = StringMap.empty;
      envParams    = StringMap.empty;
      envReturnType= constructor.ftyp;
      envBuiltinMethods = classEnv.builtinMethods;
    } in
    let _ = setEnvParams constructor.fformals env
    in 
    {
      sfaccess = constructor.faccess;
      sfname = constructor.fname;
      sfformals = List.map (fun f -> convertFormalToSast f env) constructor.fformals; 
      sftyp = constructor.ftyp;
      sfbody = List.map (fun s -> convertStmtToSast s env) constructor.fbody;
    }

  in
  (* Sematic checking for class variable *)
  let checkVdecl vdecl env = 
    "placeholder"
    (*		let check = 
      			if StringMap.mem vdecl.vname env.envClassMap.variableMap
      				then raise (Failure("Variable name already used"))
      			else if vdecl.vexpr <> Ast.Noexpr && getType vdecl.vexpr env <> vdecl.vtype
      				then raise (Failure(str_of_expr vdecl.vexpr ^ " is of type " ^ str_of_type (getType vdecl.vexpr env) ^ " but type " ^ str_of_type vdecl.vtype ^ " is expected"))
      		in check
    *)	in

  let convertVariableToSast vdecl classEnv =
    let env = {
      envClassName = classEnv.className;
      envClassMaps = classEnv.classMaps;
      envClassMap  = classEnv.classMap;
      envLocals    = StringMap.empty;
      envParams    = StringMap.empty;
      envReturnType= Void; (*@TODO make sure return is not possible here*)
      envBuiltinMethods = classEnv.builtinMethods;
    } in
    let _ = checkVdecl vdecl env
    in
    let _ =  classEnv.classMap.fieldMap <- StringMap.add vdecl.vname vdecl classEnv.classMap.fieldMap
    in {
      svaccess = vdecl.vaccess;
      svtyp  = vdecl.vtyp;
      svname  = vdecl.vname;
      svexpr  = convertExprToSast vdecl.vexpr env;
    }
  in
  let convertCbodyToSast cbody classEnv =
    {
      sfields = List.map (fun v -> convertVariableToSast v classEnv) (List.rev cbody.fields);
      sconstructors = List.map (fun cst -> convertConstructorToSast cst classEnv) (List.rev cbody.constructors);
      smethods = List.map (fun m -> convertMethodToSast m classEnv) (List.rev cbody.methods);
    }

  in
  let checkClass class_decl classEnv = 
    let firstChar = String.get class_decl.cname 0
    in
    let lowerChar = Char.lowercase firstChar
    in 
    let checking =  
      if lowerChar = firstChar
      then raise (Failure ("Class name not capitalized: " ^ class_decl.cname))
      else if StringMap.mem class_decl.cname classEnv.classMaps
      then raise (Failure ("Duplicate Class Name: " ^ class_decl.cname))
    in checking 	
  in
  let convertClassToSast class_decl classEnv =
    classEnv.className <- class_decl.cname;
    checkClass class_decl classEnv;
    let classMap = {
      fieldMap = StringMap.empty;
      constructorMap = StringMap.empty;
      methodMap = StringMap.empty;
    } in
    classEnv.classMap <- classMap;
    let result =
      {
        scaccess = class_decl.caccess;
        scname  = class_decl.cname;
        scbody  = convertCbodyToSast class_decl.cbody classEnv;
      } in
    classEnv.classMaps <- StringMap.add classEnv.className classEnv.classMap classEnv.classMaps;
    result

  in
  let classEnv = {
    className = "";
    classMaps = StringMap.empty;
    classMap  = { 
      fieldMap = StringMap.empty;
      constructorMap = StringMap.empty;
      methodMap = StringMap.empty;
    };
    builtinMethods = builtinMethods;
  }
  in
  let get_classes =
    List.map (fun c -> convertClassToSast c classEnv) classes
  in
  let sprogram = 
    {
      classes = get_classes;
      main = convertMethodToSast main classEnv;
      functions = [];
    }
  in
  sprogram

(* Translates Ast to Sast *)
let check program = match program with
    Program (classes, main) -> ignore (createClassIndices classes); convertToSast classes main

